#!/usr/bin/perl

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#	This script currently supports the following videosites:
#		[*] youtube - de, com, fr
#		[*] myvideo
#		[*] clipfish
#		[*] metacafe
#   changelog:	
#		2007-10-20	initial release
#		2007-10-20	added avi/mp3-support (encoding)
#		2007-10-20	improved code for command switches	
#				added renaming-support
#		2007-10-28	added support for myvideo & stage6
#		2007-10-29	added support for changing the videopath 
#				fixed some bugs
#		2007-11-03	added support for just printing the downloadlinks
#				realized that the code is really ugly...
#					=> TODO: rewrite the whole thing,	
#						 maybe tomorrow....
#		2007-11-08	added support for de.youtube.com
#		2008-01-30	minor changes to the code, due to youtubes 
#				latest changes....
#		2008-01-31	replaced the old commadlineparser
#		     		with Getopt::Long
#		2008-02-01	Replaced systemcalls of wget/curl with
#				LWP::UserAgent methods
#		2008-04-20	added support for clipfish, though it's
#				still a bit buggy, so don't expect it to work...
#		2008-04-21	renamed the project to
#					*yaydl - yet another youtube downloader*
#				added support for metacafe
#				removed support fpr stage6...R.I.P.
#				miscellaneous changes	 
#		2008-05-02	fixed some bugs
#		2008-06-04	fixed some minor bugs
#				changed code for myvideo due to their latest changes
#		2008-06-15	added progressbar
#		2008-06-17	added support for fr.youtube.com
#		2008-07-01	added support for beta.vreel.net
#				
#
#		If you receive an error like
#		"user@host ~ $ Can't locate LWP/UserAgent.pm in @INC (@INC contains:)..."
#		you have to install the perl module LWP/Useragent
#		On debian-based systems, simply do a  "apt-get install libwww-perl" 
#		Alternative:  use "perl -MCPAN -e shell"...
#		An older version of this script, which still uses curl/wget is also available:
#		http://home.arcor.de/haui.45/scripts/perl/yaydl_youtubedownloader/old/
#

use strict;
use warnings;
use Getopt::Long;
use LWP::UserAgent;
use MP3::Info;
use Term::ProgressBar;
use FileHandle;

Getopt::Long::Configure ("bundling", "gnu_compat");
STDOUT->autoflush;
$SIG{INT} = \&catcher;		#Catch Ctrl-C and exit

my %INFO = (
	authors => "Haui",
	contact => "haui45\@web.de",
	date	=> "2008-07-01",
	version => "v0.8.5a",
	license => "GPL",
	name	=> "yaydl - yet another youtube downloader",
);

#*************************************************************************
my $videopath = "/my/videopath/";	#IMPORTANT - change to your path!*
#*************************************************************************

my $vidid = "";		#the "url-id"
my $id = "";		#the "magic-id"
my $path = "";		#
my $encode = 0;		#output file for the encoding proces
my $source = "";	#input file for the encoding process
my $name = "";		#new filename
my $ua;			#LWP user agent
my $filehandle ;	#filehandel for syswrite

my $currentsize = 0;	#
my $bit = 0;		#
my $percent = 0;	#
my $progress;		#

my %options = (
		rename	=> '0',
		help	=> '0',
		sound	=> '0',
		version	=> '0',
		encode	=> '0',
		print	=> '0',
		tag	=> '0',
		quiet	=> '0',
);

sub catcher() {
	die "SIGINT recieved....exiting\n";
}
sub checkargs(){
	if ($#ARGV +1 == 0){
		print "Please specify at least one link...\n";
		exit;
	}
	#parse the commadline options
	GetOptions ("rename|r" => \$options{'rename'}, "help|h" => \$options{'help'},
		 "sound|s" => \$options{'sound'}, "version|v" => \$options{'version'}, 
			"encode|e" => \$options{'encode'} , "print" => \$options{'print'},
				"tag|t"=>\$options{'tag'},  "path|p=s" => \$videopath,
					"quiet|q"=>\$options{'quiet'});

	if ($options{'version'} == 1){
		version();
		$options{'version'} = 0;
		exit 0;
	}
	if ($options{'help'} == 1){
		help();
		$options{'help'} = 0;
		exit 0;
	}
	if ($#ARGV +1 == 0){
		print "No URL specified\n";
		exit;
	}
	if ($videopath =~ m/\/my\/videopath\// or $videopath eq ""){
		$videopath = $ENV{"HOME"} || 
			die "Please set a proper videopath!\n";
	}
	$videopath .= "/" unless ($videopath =~ m#/$#);
	unless (-e $videopath){
		warn "Videopath doesn't exist!\n";
		exit(1);
	}
	#print "@{[ %options ]}\n";
}


# calls the methods for the given urls
sub choose(){
	my ($path, $url,$dec);
	unless ($options{'quiet'}){
		print "\nImportant notice: All links must be quoted correctly ->" . 
			" \"http://www.youtube.com...\"\n";
		print "Type \"y\" to proceed\n";
		print "[y/n]\n";
		$dec = <STDIN>;
		chomp $dec;
		unless ($dec =~ m/^y(es)*$/i){
			warn "Quitting...\n";
			exit 1;
		}
	}
	$ua = LWP::UserAgent->new();
	$ua->agent('Mozilla/5.0');
	foreach my $link (@ARGV){
		print "---\n";
		if ($link =~ m#^http://((www\.)|(de\.)|(fr\.)){0,1}youtube\.com/watch#){
			($url, $path) = youtube($link);
		}
		elsif ($link =~ m#^http://(www\.){0,1}myvideo\.de\/watch#){
			($url, $path) = myvideo($link);
		}
		elsif (($link =~ m#^http://(www\.){0,1}clipfish\.de/player\.php\?videoid=.*#)){
			($url, $path) = clipfish($link);
		}
		elsif (($link =~ m#^http://www.metacafe.com/watch/.*#)){
			($url, $path) = metacafe($link);
		}
		elsif (($link =~ m#^http://beta.vreel.net/watch_.*#)){
			($url, $path) = vreel($link);
		}
		else{
			warn "$link: Not a valid Link!\n";
			next;
		}
		if ($url eq ""){
			next;
		}
		#just print the downloadlink...
		if($options{'print'}){
			print "Downloadlink: $url\n";
			next;
		}
			
		warn "Download in progress - this may take a while\n";
		if (download($url, $path) != 0){
			die "Error while getting the file!\n\n";
		}
		warn "\n[***************]Saved file $path [***************]\n\n";

		if($options{'encode'} == 1){
			if (encode("$path") == -1){

			}
			else{
			warn "\n[**************] Encoded file $encode [****************]\n\n" unless 
				($options{'rename'} == 1);
			}
		}
		if($options{'sound'} == 1){
			if (sound("$path") == -1){
				die "ripping error!\n";
			}
			warn "\n[**************] Encoded file $encode [****************]\n\n" unless 
				($options{'rename'} == 1);
		}
		if ($options{'sound'} == 1 || $options{'encode'} == 1 ){
			unlink("$path");
		}
		else{
			if ($options{'rename'} == 1){
				print "\nEnter a new filename\n";
				$name = <STDIN>;
				chomp($name);
				$name = "$videopath" . "/$name";
				rename("$path", "$name") or
					die "renaming failed!\n\n";
				warn "\n$path renamed successfully to $name\n\n";
		}
		$path="";
		}

	}
		print "---\n";
}

sub youtube(){
	my $url = shift;
	my $line = "http://youtube.com/get_video?video_id=" ;
	$vidid = $url;
	$vidid =~ s#http://((de\.)|(www\.)|(fr\.))*youtube.com/watch\?v=##;
	$vidid =~ s#\&.*##;
	$path = "$videopath" . "/$vidid" . ".flv";
	if ((getpage($url)) != 0){
		warn "Error while getting the file!\n\n";
		return "";
	}
	my @testing = split(/\n/, $id);
	my @testme = grep (/^.*, "t": ".*", "hl":.*/, @testing);
	unless(defined $testme[0]){
		warn "Error while getting the file!\n\n";
		return "";
	}
	$id = $testme[0];
	$id =~ s/^.*, "t": "//;		#alles vor und inkl. ', "t": "' loeschen
	$id =~ s/", "hl":.*//;		#alles nach und inkl. '", "hl":' loeschen
	$url = $line . $vidid . "\&t=" . $id;	# url zusammenfuegen
	return $url, $path;
}

sub vreel {
	my $url = shift;
	if ((getpage($url)) != 0){
		warn "Error while getting the file!\n\n";
		return "";
	}
	$url =~ s#(.*watch_)(.*)(\.html)#$2#; # get the videonumber
	$path = $videopath . $url . ".flv";

	# extract the downloadlink from the http-page
	my @testing = split(/\n/, $id);
	my @testme = grep (/^.*embed type=.*llnwd.*\.avi.*/, @testing);
	unless(defined $testme[0]){
		warn "Error while getting the file!\n\n";
		return "";
	}
	$id = $testme[0];
	$id =~ s#.*src="##;	# get the downloadlink for the video
	$id =~ s#" .*##;	#
	chomp $id;
	$url = $id;
	return $url, $path;

}

sub clipfish(){
	my ($url, $serverid) ="";
	$url = shift;
	my $server="http://pg1.clipfish.de/media/";
	if ((getpage($url)) != 0){
		print "Error while getting the file!\n\n";
		return "";
	}
	
	#get the id for the flv file	
	my @array = split(/\n/, $id);
	my @ids = grep (/\s*<!--.*flv.*/, @array);
	unless (defined $ids[0]){
		print "ids[0] undefined...stopping method\n\n";
		return "";
	}
	my @ismusic = grep (/media\/musicvideos/, @array);

	$vidid = $ids[0];
	chomp($vidid);
	$vidid =~ s/.*<!--//;
	$vidid =~ s/\.flv.*/\.flv/;
	#set the download path
	$path = $videopath . $vidid;
	
	#music videos need a special treatment...
	if (defined $ismusic[0]) {
		print "It's a  musicvideo! :D\n ";
		$url = $server . "musicvideos/" . substr($vidid, 0, 2) . 
		"/" . substr($vidid,2,3) . "/" .  substr($vidid, 5,2 ) . "/" . $vidid;

	}
	else {
		#get the two-digit id for the download server	
		$serverid = $vidid;
		$serverid =~ s/\.flv//;
		$serverid = substr($serverid, length($serverid)-2, 2);
	
		#create the url....
		$url = $server . $serverid . "/" . $vidid;
	}	
	return $url, $path;
}
sub metacafe(){
	my $url = shift;
	my $path = $url;
	if ((getpage($url)) != 0){
		print "Error while getting the file!\n\n";
		return "";
	}
	my @array = split("\n", $id);
	my @ids = grep (/so\.addParam\("flashvars", "itemID=/, @array);
	unless (defined $ids[0]){
		print "Error while getting the file!\n\n";
		return "";
	}
	$url = $ids[0];
	$url =~ s#.*mediaURL=##;
	$url =~ s#\.flv.*#\.flv#;
	
	$path =~ s#http://www.metacafe.com/watch/##;
	$path =~ s#/.*##;
	$path = $videopath . $path . ".flv";
	return $url, $path;
}

sub myvideo(){
	my ($path, $url);
	my @swfargs;
	$vidid = shift;

	if ((getpage($vidid)) != 0){
		print "Error while getting the file!\n\n";
		return "";
	}
		
	my @array = split(/\n/, $id);

	$vidid =~ s#(http://)?(www\.)?myvideo\.de/watch/##;
	$vidid =~ s#/.*##;
	$path = "$videopath" . "/$vidid" . ".flv";

	@swfargs = grep /SWFObject\(/, @array;
	unless (defined($swfargs[0])) {
		print "variable swfargs undefined...stopping method\n\n"; 
		return "";
	}
	$id=$swfargs[0];
	undef (@array);
	undef (@swfargs);

	$id =~ s#.*SWFObject\('##;
	$id =~ s#;.*##;
	$id =~ s#.*\?##;
	$id =~ s#\&.*##;
	chomp($id);
	$url = $id;
	return $url, $path;

}

sub download() {
	my ($url, $path) = @_;
	print "Downloalink: $url\n";
	print "File: $path\n\n";
	my ($request, $response);
	my $chunksize=1024;
	$request = HTTP::Request->new('HEAD', $url);			#	Filesize
	$response = $ua->request($request);				# 	.
	if ($response->is_success()){					#	.
		if (int($response->header('Content-Length')) == 0){
			print "Filesize 0 byte....stopping method!\n\n";
			return -1;
		}
		print "Filesize: " . int($response->header('Content-Length')/1024) . "kb\n"; #
	}								#	.
	else{								#	.
		print "Error code    : ", $response->code() ,    "\n";	#	.
		print "Error message:  ", $response->message() , "\n";	#	.
		return -1;						#	.
	}								#	.
	$bit = int ($response->header('Content-Length')/100); 
	$request = HTTP::Request->new('GET', $url);			#
	open $filehandle, '>', $path;
 	$progress = Term::ProgressBar->new (100);
	$response = $ua->request($request, \&callback, $chunksize);

	if ( $response->is_error() ) { 					#	
	       print "Error code    : ", $response->code() ,    "\n";	#
		print "Error message:  ", $response->message() , "\n";	#
		return 1;						#
	}
	close $filehandle;	
	$percent = 0;
	$currentsize = 0;
	$bit = 0;
	undef $progress;
	undef $filehandle;
	print "\n";


	return 0;
	
}

sub callback {
	my $data = shift;
	my $length = length $data;
	$currentsize += $length;
	if($currentsize >= $percent*$bit){
		$percent++;
		$progress->update ($percent-1);
	}
	syswrite $filehandle, $data, $length; 
}


sub getpage() {
	my $link = shift;

	my ($path, $request, $response, $url) = "";
	$request = HTTP::Request->new('GET', $link);
	$response = $ua->request($request);
	print "Retrieving http-page...\n";
	if ( $response->is_error() ) { 
		print "Error code    : ", $response->code() ,    "\n";
		print "Error message:  ", $response->message() , "\n";
		return -1;
	}
	else {
		$id = $response->content();
		return 0;
	}

}

sub encode(){
	$source = shift;
	$encode = $source;
	$encode =~ s/\.flv/\.avi/;
	$encode =~ s/\.divx/\.avi/;
	my $ffmpeg = `which ffmpeg`;
	my $mencoder = `which mencoder`;
	my @commands;
	my $return;
	if ($mencoder ne ""){
		 push(@commands, "mencoder \"$source\" -oac mp3lame -lameopts abr:br=92" .
		" -ovc xvid -xvidencopts bitrate=150 -o \"$encode\"");
		print "mencoder: found\n";
	}
	if($ffmpeg ne ""){
		push (@commands, "ffmpeg -i \"$source\" \"$encode\"");
		print "ffmpeg: found\n";
	}
	if($ffmpeg eq "" && $mencoder eq ""){
		$options{'encode'} = 0;
		print "mencoder: not found!\nffmpeg: not found!\nencoding disabled\n";
		return -1;
	}
	foreach my $command(@commands){
		$return = system($command);
		if ($return == 0){
			print "success\n";
			last;
		}
	}
	if ($return != 0){
		print "error - encoding disabled\n";
		$options{'encode'} = 0;
		return -1;
	}


	if ($options{'rename'} == 1){
		print "\nEnter a new filename\n";
		$name = <STDIN>;
		chomp($name);
		$name =~ s/.avi$//;
		$name = "$videopath" . "/$name" . ".avi";
		rename("$encode", "$name") or die "encode rename: Something went wrong. Fix it! :)\n";
		print "\n$encode renamed successfully to $name\n\n";
	}

}

sub sound(){
	$source = shift;
	$encode = $source;
	$encode =~ s/\.flv/\.mp3/;
	$encode =~ s/\.divx$/\.mp3/;
	system("mplayer -dumpaudio -dumpfile \"$encode\" \"$source\"") == 0
			or return -1;
	if ($options{'rename'} == 1){
		print "\nEnter a new filename\n";
		$name = <STDIN>;
		chomp($name);
		$name =~ s/.mp3$//;
		$name = "$videopath" . "/$name" . ".mp3";
		rename("$encode", "$name") or die "sound rename: Something went wrong. Fix it!\n";
		print "\n$encode renamed successfully to $name\n\n";
	}
	if ($options{'tag'} == 1){
		eval {
			print "Enter new title: ";
			my $title = <STDIN>;
			print "Enter new artist: ";
			my $artist = <STDIN>;
			print "Enter new album: ";
			my $album = <STDIN>;
			chomp($title, $artist, $album);
			my $mp3 = new MP3::Info "$name";
			$mp3->title($title);
			$mp3->artist($artist);
			$mp3->album($album);
			print "tagging successful\n\n";
		};
		if($@){
			print "tagging error\n";
			return -1;
		}

	}
	return 0;
}



sub help(){
	print "Usage: yaydl  [-esrtphv] \"link1\" \"link2\" ...\n".

	 "\nImportant notice: all links must be quoted correctly!\n".
	 "Parameters:\n".
	 "   -e, --encode \t Encode the videos from *.flv to *.avi - requires mencoder or ffmpeg \n".
	 "    \t\t\t yaydl -e \"http://youtube...\"\n".
	 "   -s, --sound\t\t Extract the soundtracks of the downloaded videos - requires mplayer!\n".
	 "    \t\t\t yaydl -s \"http://youtube...\"\n".
	 "   -r, --rename \t Rename the downloaded files\n" .
	 "   -t, --tag\t\t Provides a basic way for tagging the extracted soundfiles\n" .
	 "   -p, --path PATH \t Change the default downloadpath to PATH\n" . 
	 "       --print \t\t Display the downloadlinks and exit\n" . 
	 "   -q, --quiet\t Don't wait for the user's confirmation\n".
	 "   -h, --help   \t Show this help.\n".
	 "   -v, --version \t Print version.\n";
	
	return;
}
sub version(){
	print "yaydl - yet another youtube downloader \nversion $INFO{'version'} - $INFO{'date'}\n";
	print "Author: $INFO{'authors'}, contact $INFO{'contact'}\n";
	return;
}

checkargs();
choose();

